import re
import sys
from datetime import datetime

from ._quinlan_attributes import quinlan_attributes


def make_names_string(x, w=None, label="outcome"):
    """
    Create the names string to pass to Cubist. This string contains information about Python and the time of run along
    with the column names and their data types.

    Parameters
    ----------
    x : {pd.DataFrame} of shape (n_samples, n_features)
        The input samples.

    w : ndarray of shape (n_samples,)
        Instance weights.
    
    label : str, default="outcome"
        A label for the outcome variable. This is only used for printing rules.

    Returns
    -------
    out : str
        Case name string describing training dataset columns and their types.
    """
    # copy Pandas objects so they aren't changed outside of this function
    x = x.copy(deep=True)

    # clean reserved sample name if it's in x
    has_sample = [i for i, c in enumerate(x.columns) if bool(re.search('^sample', c))]
    if has_sample:
        x.columns = [re.sub('^sample', '_Sample', c) for c in x.columns]

    # generate the comments string showing the Python version and current timestamps
    python_version = tuple(sys.version_info)
    now = datetime.now()
    out = f'| Generated using Python {python_version[0]}.{python_version[1]}.{python_version[2]}\n' \
          f'| on {now.strftime("%a %b %d %H:%M:%S %Y")}'

    # define the outcome data type
    outcome_type = ": continuous."

    # build base out string
    out = f'{out}\n{label}.\n{label}{outcome_type}'

    # get dictionary of feature names as keys and data types as values
    var_data = quinlan_attributes(x)

    # if weights are present add this to var_data
    if w is not None:
        var_data["case weight"] = "continuous."

    # join the column names and data types into a single string
    var_data = [f'{escapes([key])[0]}: {value}' for key, value in var_data.items()]
    var_data = '\n'.join(var_data)

    # merge the out and var_data strings
    out = f'{out}\n{var_data}\n'
    return out


def escapes(x, chars=None):
    """Double escape reserved and special characters in x."""
    # set custom reserved characters list
    if chars is None:
        chars = [':', ';', '|']
    # apply first escaping
    for i in chars:
        x = [c.replace(i, f'\\{i}') for c in x]
    # apply second escaping
    x = [re_escape(c) for c in x]
    return x


_special_chars_map = {i: '\\' + chr(i) for i in b'()[]{}?*+-|:;^$\\.&~#\t\n\r\v\f'}


def re_escape(pattern):
    """Escape special characters in a string. Sourced from 're' Python package."""
    if isinstance(pattern, str):
        return pattern.translate(_special_chars_map)
    else:
        pattern = str(pattern, 'latin1')
        return pattern.translate(_special_chars_map).encode('latin1')



import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype, is_datetime64_any_dtype


def _get_data_format(x: pd.Series):
    """
    Function to obtain the data type/formatting information for a Pandas Series.
    Return "continuous." for continuous features, the set of values as a comma
    separated string for categorical features, and the column itself for 
    datetime features.

    Parameters
    ----------
    x : pd.Series
        Pandas Series from which to extract data type.
    
    Returns
    -------
    x : str
        String description of the Series data type.
    """
    # remove NAs from series
    x = x.dropna()
    # for numeric columns
    if is_numeric_dtype(x):
        return "continuous."
    # for string columns
    if is_string_dtype(x):
        return f"{','.join(set(x))}."
    # for datetime columns
    if is_datetime64_any_dtype(x):
        return x
    # otherwise the data is not supported here
    else:
        raise NotImplementedError


def quinlan_attributes(df: pd.DataFrame) -> dict:
    """
    Function to collect the data formatting information for each column in a Pandas DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame from which column data attributes are obtained.
    
    Returns
    -------
    x : dict
        Dictionary with keys as column names and values as the description of
        the data type.
    """
    return {col_name: _get_data_format(col_data) for col_name, col_data in df.iteritems()}



import warnings

import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype
import numpy as np

from ._make_names_string import escapes


def validate_x(x):
    """Ensure input dataset is of a valid type and format"""
    assert isinstance(x, (pd.DataFrame, np.ndarray)), "X must be a Numpy Array or Pandas DataFrame"
    if isinstance(x, np.ndarray):
        assert len(x.shape) == 2, "Input NumPy array has more than two dimensions, only a two dimensional matrix " \
                                  "may be passed."
        warnings.warn("Input data is a NumPy Array, setting column names to default `var0, var1,...`.")
        x = pd.DataFrame(x, columns=[f'var{i}' for i in range(x.shape[1])])
    return x


def r_format(x: float, digits: int = 15) -> str:
    """Python version of the R format function to return a number formatted as a 
    string rounded to `digits` number of digits from the left."""
    # if x is NA return NA
    if pd.isna(x):
        return x
    # get the count of whole number digits, i.e. the number of digits to the left of the decimal place
    whole_nums_count = len(str(int(x)))
    # Where there are decimal places that need to be rounded, round to digits - whole_nums_count decimal places
    if whole_nums_count < digits:
        remaining_decimals = digits - whole_nums_count
        return str(round(x, remaining_decimals))
    # Where there are no decimals that need to/can be rounded
    else:
        return str(x)


def make_data_string(x, y=None, w=None):
    """
    Converts input dataset array X into a string.

    Parameters
    ----------
    x : {pd.DataFrame} of shape (n_samples, n_features)
        The input samples.

    y : pd.Series
        The predicted values.
    
    w : ndarray of shape (n_samples,)
        Instance weights.

    Returns
    -------
    x : str
        Input dataset converted to a string and formatted per Cubist's requirements.
    """
    # copy Pandas objects so they aren't changed outside of this function
    x = x.copy(deep=True)
    y = y.copy(deep=True)

    # apply the escapes function to all string columns
    for col in x:
        if is_string_dtype(x[col]):
            x[col] = escapes(x[col].astype(str))

    # if y is None for model predictions, set y as a column of NaN values, which will become ?'s later
    if y is None:
        y = [np.nan] * x.shape[0]
        y = pd.Series(y)

    # format the y column for special charactesrs
    y = pd.Series(escapes(y.astype(str)))

    # insert the y column as the first column of x
    x.insert(0, "y", y)

    # handle weights matrix (?) TODO: validate
    if w is not None:
        column_names = list(x.columns) + [f"w{i}" for i in range(w.shape[1])]
        x = pd.concat([x, w], axis=1, ignore_index=True)
        x.columns = column_names

    # convert all columns to strings
    for col in x:
        if is_numeric_dtype(x[col]):
            x[col] = x[col].apply(r_format)
            x[col] = x[col].astype(str)
        else:
            x[col] = x[col].astype(str)

    # remove leading whitespace from all elements
    x = x.applymap(lambda a: a.lstrip())

    # replace missing values with ?
    x = x.fillna("?")
    x = x.replace("nan", "?")

    # convert dataframe to list of lists
    x = x.to_numpy().tolist()

    # merge each sublist into single strings with entries separated by commas
    x = [','.join(row) for row in x]

    # join all row strings into a single string separated by \n's
    x = "\n".join(x)
    return x



import re

import pandas as pd
import numpy as np


def count_rules(x):
    return


def split_to_groups(x, f):
    """Function to convert two lists into a dictionary where the keys are unique values in f and 
    the values are lists of the corresponding values in x. Analogous to the split function in R."""
    if len(x) != len(f):
        raise ValueError("lists x and f must be of the same length")
    groups = {}
    for a, b in zip(x, f):
        if b in groups:
            groups[b].append(a)
        else:
            groups[b] = [a]
    return groups


def parse_cubist_model(model, x):
    print(model)
    # split on newline
    model = model.split("\n")
    # remove empty strings
    model = [c for c in model if c.strip() != '']
    # get model length
    model_len = len(model)

    # define initial lists and index variables
    com_num = [None] * model_len
    rule_num = [None] * model_len
    cond_num = [None] * model_len
    com_idx = r_idx = 0

    # loop through model and indicate
    for i, row in enumerate(model):
        # break each row of x into dicts for each key/value pair
        tt = parser(row)

        # get the first key in the first entry of tt
        first_key = list(tt[0].keys())[0]

        # start of a new rule
        if first_key == "rules":
            com_idx += 1
            r_idx = 0
        com_num[i] = com_idx
        # start of a new condition
        if first_key == "conds":
            r_idx += 1
            c_idx = 0
        rule_num[i] = r_idx
        # within a rule, type designates the type of conditional statement
        # type = 2 appears to be a simple split of a continuous predictor
        if first_key == "type":
            c_idx += 1
            cond_num[i] = c_idx
    
    # get the number of committees
    num_com = len([c for c in model if re.search("^rules=", c)])

    # apply the analogous R split function to get a dict
    rules_per_com = split_to_groups(rule_num, com_num)

    rules_per_com = {a: max(rules_per_com[a]) for a in rules_per_com}
    rules_per_com = {a: rules_per_com[a] for a in rules_per_com if rules_per_com[a] > 0}
    if rules_per_com and num_com > 0:
        rules_per_com = {f'Com {i + 1}': rules_per_com[a] for i, a in enumerate(list(rules_per_com.keys()))}
    
    split_var = [None] * model_len
    split_val = [None] * model_len
    split_cats = [""] * model_len
    split_dir = [""] * model_len
    split_type = [""] * model_len

    is_type2 = [i for i, c in enumerate(model) if re.search("^type=\"2\"", c)]
    for i in is_type2:
        split_type[i] = "type2"
        continuous_split = type2(model[i])
        split_var[i] = continuous_split["var"].replace('\"', "")
        split_dir[i] = continuous_split["rslt"]
        split_val[i] = continuous_split["val"]

    is_type3 = [i for i, c in enumerate(model) if re.search("^type=\"3\"", c)]
    for i in is_type3:
        categorical_split = type3(model[i])
        split_var[i] = categorical_split["var"]
        split_cats[i] = categorical_split["val"]

    if is_type2 == [] and is_type3 == []:
        return None

    split_data = pd.DataFrame({
        "committee": com_num,
        "rule": rule_num,
        "variable": split_var,
        "dir": split_dir,
        "value": split_val,
        "category": split_cats,
        "type": split_type
    })
    split_data = split_data.dropna(subset=['variable'])
    split_data = split_data.reset_index(drop=True)

    # get the rule by rule percentiles (?)
    nrows = x.shape[0]
    for i in range(split_data.shape[0]):
        var = split_data.loc[i, "value"]
        if not np.isnan(var):
            x_col = pd.to_numeric(x[split_data.loc[i, "variable"]])
            split_data.loc[i, "percentile"] = sum([c <= var for c in x_col]) / nrows
    
    # get coefficients
    is_eqn = [i for i, c in enumerate(model) if "coeff=" in c]
    coefs = [eqn(model[i], var_names=list(x.columns)) for i in is_eqn]
    out = pd.DataFrame(coefs)
    out["committee"] = [com_num[i] for i in is_eqn]
    out["rule"] = [rule_num[i] for i in is_eqn]

    # get maxd
    tmp = [c for c in model if "maxd" in c][0]
    tmp = tmp.split("\"")
    maxd_i = [i for i, c in enumerate(tmp) if "maxd" in c][0]
    maxd = tmp[maxd_i + 1]
    print(split_data, out, maxd)
    return split_data, out, float(maxd)


def type3(x):
    a_ind = x.find("att=")
    e_ind = x.find("elts=")
    var = x[a_ind + 5:e_ind - 2]
    val = x[e_ind + 5:]
    val = val.replace("[{}]", "").replace("\"", "").replace(" ", "")
    mult_vals = "," in val
    val = val.replace(",", ", ")
    if mult_vals:
        val = f"{{mult_vals}}"
        txt = f"{var} in {val}"
    else:
        txt = f"{var} = {val}"
    return {"var": var, "val": val, "text": txt}


def type2(x, dig=3):
    x = x.replace("\"", "")
    a_ind = x.find("att=")
    c_ind = x.find("cut=")
    r_ind = x.find("result=")
    v_ind = x.find("val=")

    missing_rule = c_ind < 1 and v_ind > 0
    if missing_rule:
        var = x[a_ind + 4:c_ind - 1]
        val = None
        rslt = "="
    else:
        var = x[a_ind + 4:c_ind - 1]
        val = x[c_ind + 4:r_ind - 1]
        val = round(float(val), dig)
        rslt = x[r_ind + 7:]
    return {"var": var,
            "val": val,
            "rslt": rslt,
            "text": f"{var} {rslt} {val}"}


def eqn(x, var_names=None):
    x = x.replace("\"", "")
    starts = [m.start(0) for m in re.finditer("(coeff=)|(att=)", x)]
    tmp = [""] * len(starts)
    for i in range(len(starts)):
        if i < len(starts) - 1:
            txt = x[starts[i]:starts[i + 1] - 1]
        else:
            txt = x[starts[i]:]
        tmp[i] = txt.replace("coeff=", "").replace("att=", "")

    vals = tmp[::2]
    vals = [float(c) for c in vals]
    nms = tmp[1::2]
    nms = ["(Intercept)"] + nms
    vals = {nm: val for nm, val in zip(nms, vals)}
    if var_names:
        vars2 = [var for var in var_names if var not in nms]
        vals2 = [np.nan] * len(vars2)
        vals2 = {nm: val for nm, val in zip(vars2, vals2)}
        vals = {**vals, **vals2}
        new_names = ["(Intercept)"] + var_names
        vals = {nm: vals[nm] for nm in new_names}
    return vals


def make_parsed_dict(x):
    """"""
    x = x.split("=")
    if len(x) > 1:
        return {x[0]: x[1]}
    else:
        return None


def parser(x):
    """"""
    x = x.split(" ")
    x = [make_parsed_dict(c) for c in x]
    return x



from random import randint

import numpy as np
import pandas as pd
from sklearn.base import RegressorMixin, BaseEstimator

from ._make_names_string import make_names_string
from ._make_data_string import make_data_string, validate_x
from ._parse_cubist_model import parse_cubist_model
from ._variable_usage import get_variable_usage
from _cubist import _cubist, _predictions


class Cubist(RegressorMixin, BaseEstimator):
    """
    Cubist Regression Model (Public v2.07) developed by JR Quinlan.

    References: 
    - https://www.rdocumentation.org/packages/Cubist/versions/0.3.0
    - https://www.rulequest.com/cubist-unix.html


    Parameters
    ----------
    n_rules : int, default=100
        Limit of the number of rules Cubist will build.

    n_committees : int, default=1
        Number of committees to construct. Each committee is a rule based model and beyond the first tries to correct the prediction errors of the prior constructed model.

    unbiased : bool, default=False
        Should unbiased rules be used? Since Cubist minimizes the MAE of the predicted values, the rules may be biased and the mean predicted value may differ from the actual mean. This is recommended when there are frequent occurrences of the same value in a training dataset. Note that MAE may be slightly higher.

    extrapolation : float, default=1.0
        Adjusts how much rule predictions are adjusted to be consistent with the training dataset.

    sample : float, default=0.0
        Percentage of the data set to be randomly selected for model building.

    seed : int, default=randint(0, 4095)
        An integer to set the random seed for the C Cubist code.

    target_label : str, default="outcome"
        A label for the outcome variable. This is only used for printing rules.

    weights : array-like of shape (n_samples,)
        Optional vector of case weights that is the same length as y for how much each instance should contribute to the model fit.

    neighbors : int, default=0
        Number between 0 and 9 for how many instances should be used to correct the rule-based prediction.

    verbose : bool, default=False
        Should the Cubist output be printed?

    Attributes
    ----------
    names_string : str
        String for the Cubist model that describes the training dataset column names and their data types. This also provides some Python environment information.

    model : str
        The Cubist model string generated by the C code.

    maxd : float

    variable_usage : pd.DataFrame
        Table of how training data variables are used in the Cubist model.

    rules : pd.DataFrame
        Table of the rules built by the Cubist model.

    coefficients : pd.DataFrame
        Table of the regression coefficients found by the Cubist model.

    variables : dict
        Information about all the variables passed to the model and those that were actually used.

    Examples
    --------
    >>> from cubist import Cubist
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.model_selection import train_test_split
    >>> X, y = load_boston(return_X_y=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    >>> model = Cubist()
    >>> model.fit(X_train, y_train)
    >>> model.predict(X_test)
    >>> model.score(X_test, y_test)
    """
    def __init__(self,
                 n_rules: int = 100,
                 n_committees: int = 1,
                 unbiased: bool = False,
                 extrapolation: float = 1.0,
                 sample: float = 0.0,
                 seed: int = randint(0, 4095),  # TODO fix this since its different from R
                 target_label: str = "outcome",
                 weights=None,
                 neighbors: int = 0,
                 verbose: bool = False):
        super().__init__()

        # initialize instance variables
        self.n_rules = n_rules
        self.n_committees = n_committees
        self.unbiased = unbiased
        self.extrapolation = extrapolation
        self.sample = sample
        self.seed = seed
        self.target_label = target_label
        self.weights = weights
        self.neighbors = neighbors
        self.verbose = verbose

        # initialize remaining instance variables
        self.names_string = None
        self.model = None
        self.maxd = None
        self.variable_usage = None
        self.rules = None
        self.coefficients = None
        self.variables = None

    @property
    def n_rules(self):
        return self._n_rules

    @n_rules.setter
    def n_rules(self, value):
        assert value is not None and (value > 1 or value < 1000000), "number of rules must be between 1 and 1000000"
        self._n_rules = value
    
    @property
    def n_committees(self):
        return self._n_committees

    @n_committees.setter
    def n_committees(self, value):
        assert value > 1 or value < 100, "number of committees must be between 1 and 100"
        self._n_committees = value

    @property
    def extrapolation(self):
        return self._extrapolation

    @extrapolation.setter
    def extrapolation(self, value):
        assert 0.0 <= value <= 1.0, "extrapolation percentage must be between 0.0 and 1.0"
        self._extrapolation = value

    @property
    def sample(self):
        return self._sample

    @sample.setter
    def sample(self, value):
        assert 0.0 <= value <= 1.0, "sampling percentage must be between 0.0 and 1.0"
        self._sample = value

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, value):
        self._seed = value % 4095

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, value):
        assert value is None or isinstance(value, (list, np.ndarray)), "case weights must be numeric"
        self._weights = value

    @property
    def neighbors(self):
        return self._neighbors

    @neighbors.setter
    def neighbors(self, value):
        assert isinstance(value, int), "Only an integer value for neighbors is allowed"
        assert 10 >= value >= 0, "'neighbors' must be 0 or greater and 10 or less"
        self._neighbors = value

    def fit(self, X, y):
        """
        Build a Cubist regression model from training set (X, y).

        Parameters
        ----------
        X : {array-like} of shape (n_samples, n_features)
            The training input samples.

        y : array-like of shape (n_samples,)
            The target values (Real numbers in regression).

        Returns
        -------
        self : object
        """
        assert isinstance(y, (list, pd.Series, np.ndarray)), "Cubist requires a numeric target outcome"
        if not isinstance(y, pd.Series):
            y = pd.Series(y)

        # validate input data
        # X, y = self._validate_data(X, y)
        X = validate_x(X)
        X = X.reset_index(drop=True)
        y = y.reset_index(drop=True)
        X_columns = list(X.columns)

        # create the names and data strings required for cubist
        self.names_string = make_names_string(X, w=self.weights, label=self.target_label)
        data_string = make_data_string(X, y, w=self.weights)

        # call the C implementation of cubist
        self.model, output = _cubist(self.names_string.encode(),
                                     data_string.encode(),
                                     self.unbiased,
                                     b"yes",
                                     1,
                                     self.n_committees,
                                     self.sample,
                                     self.seed,
                                     self.n_rules,
                                     self.extrapolation,
                                     b"1",
                                     b"1")

        # convert output from raw to strings
        self.model = self.model.decode()
        output = output.decode()

        # replace "__Sample" with "sample" if this is used in the model
        if "\n__Sample" in self.names_string:
            output = output.replace("__Sample", "sample")
            self.model = self.model.replace("__Sample", "sample")
            # clean model string to fix breaking predictions when using reserved sample name
            self.model = self.model[:self.model.index("sample")] + self.model[self.model.index("entries"):]

        # raise cubist errors
        if "Error limit exceeded" in output:
            raise Exception(output)

        # print model output if using verbose output
        if self.verbose:
            print(output)

        # get dataframes containing the rules and coefficients
        self.rules, self.coefficients, self.maxd = parse_cubist_model(self.model, X)

        # extract the maxd value and remove the preceding text from the model string
        self.model = self.model.replace(f'insts=\"1\" nn=\"1\" maxd=\"{int(self.maxd)}\"', 'insts=\"0\"')

        # get the input data variable usage
        self.variable_usage = get_variable_usage(output, X)

        # get the names of columns that have no nan values
        not_na_cols = self.coefficients.columns[~self.coefficients.isna().any()].tolist()

        # skip the first three since these are always filled
        not_na_cols = not_na_cols[3:]

        # store a dictionary containing all the training dataset columns and those that were used by the model
        if self.rules is not None:
            used_variables = set(self.rules["variable"]).union(
                set(not_na_cols)
            )
            self.variables = {"all": X_columns,
                              "used": list(used_variables)}
        
        return self

    def predict(self, X):
        """
        Predict Cubist regression target for X.

        Parameters
        ----------
        X : {array-like} of shape (n_samples, n_features)
            The input samples. 

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted values.
        """
        # validate input data
        X = validate_x(X)
        X = X.reset_index(drop=True)

        if self.neighbors > 0:
            model = self.model.replace("insts=\"0\"", f"insts=\"1\" nn=\"{self.neighbors}\" maxd=\"{self.maxd}\"")
        else:
            model = self.model

        # If there are case weights used during training, the C code will expect a column of weights in the new data but the
        # values will be ignored. `makeDataFile` puts those last in the data when `cubist.default` is run, so we will add a
        # column of NaN values at the end here
        if self.weights:
            X["case_weight_pred"] = np.nan

        # make data string for predictions
        data_string = make_data_string(X)

        # get cubist predictions from trained model
        pred, output = _predictions(data_string.encode(),
                                    self.names_string.encode(),
                                    model.encode(),
                                    np.zeros(X.shape[0]),
                                    b"1")
        
        # TODO: parse and handle errors in output
        if output:
            print(output)
        return pred
